////////////////////////////////////////////////////////////////////////////
// Filename:    LoadXmlFile.cpp
// Purpose:     Loads an Do178Builder XML file.
// Mods:        09/23/01 RSB    Began.
//              10/13/01 RSB    Added traceability.
//              10/21/01 RSB    Added control of output directory.
//                              Now ignores TBD entries in document 
//                              production.  Parses and includes the header
//                              info now.
//              10/22/01 RSB    Debugged printout of multi-paragraph text.
//                              Eliminated the automatic removal of "TBD"
//                              sections, since if those are auto-eliminated
//                              it tends to give the false impression that
//                              the documents are more complete than they
//                              really are.
//              10/29/01 RSB    Now allow the DocBook tags <funcname>,
//                              <type>, <varname>, and <programlisting>
//                              to be written to the sgml files.
//                              Fixed bug in which &#nn; misinterpreted
//                              on input.
//              10/30/01 RSB    Added DocBook <emphasis> and 
//                              </emphasis> and a lot of others, too.
//              10/31/01 RSB    Now recognizes all DocBook 3.1 inline tags.
//                              Added backup files on save.
//                              Fixed serious bug in which TYPE is saved with
//                              spaces rather than underlines, and hence is
//                              discarded on later loading.
//              02/17/02 RSB    Corrected interpretation of '&', '<', '>' in
//				the input XML file.

/*
  Copyright 2001,2002 Ronald S. Burkey

  This file is part of Do178Builder.

  Do178Builder is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Do178Builder is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Do178Builder; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

////////////////////////////////////////////////////////////////////////////

// For compilers that support precompilation
#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

// Include private headers
#include "Do178Builder.h"

//------------------------------------------------------------------
// Clears an XML list in memory.  All linked records and their
// linked data are deallocated.  The root of the linked list, however,
// is merely cleared, and no attempt is made to deallocate it.

void
ClearXml (struct XmlStruct *Xml)
{
  struct XmlStruct *Next, *NextNext;
  for (Next = Xml; Next != NULL;)
    {
      if (Next->Children != NULL)
	{
	  ClearXml (Next->Children);
	  // Remember, ClearXml won't deallocate the root of 
	  // Next->Children, so we have to do that explicitly.
	  free (Next->Children);
	  Next->Children = NULL;
	}
      if (NULL != Next->Name)
	{
	  free (Next->Name);
	  Next->Name = NULL;
	}
      if (NULL != Next->Hint)
	{
	  free (Next->Hint);
	  Next->Hint = NULL;
	}
      if (NULL != Next->Text)
	{
	  free (Next->Text);
	  Next->Text = NULL;
	}
      // Deallocate this entire record -- unless it's the root. 
      NextNext = Next->Next;
      if (Next != Xml)
	free (Next);
      else
	Next->Next = NULL;
      Next = NextNext;
    }
}

//------------------------------------------------------------------
// Converts a tag name to something nicely displayable, by converting
// underlines to spaces, and some other conversions as well.

static void
TagToName (char *Text)
{
  char c, *s;
  int i;
  while (*Text)
    {
      if (*Text == '_')
	*Text++ = ' ';
      else if (*Text == '&' && 2 == sscanf (Text, "&#%d%c", &i, &c)
	       && c == ';')
	{
	  if (i == '\n' || (i >= ' ' && i < 127))
	    *Text++ = i;
	  else
	    *Text++ = '.';
	  for (s = Text; *s != ';'; s++);
	  strcpy (Text, s + 1);
	}
      else
	Text++;
    }
}

// ... and vice-versa.

void
NameToTag (const char *Name, char *Tag)
{
  while (*Name)
    {
      if (*Name == ' ')
	{
	  *Tag++ = '_';
	  Name++;
	}
      else if (*Name == '-' || isalnum (*Name))
	*Tag++ = *Name++;
      else
	{
	  sprintf (Tag, "&#%d;", *Name++);
	  Tag += strlen (Tag);
	}
    }
  *Tag = 0;
}

//------------------------------------------------------------------
// Reads a file until a markup tag is found, and returns the tag.
// The return value is 0 on success.  Call the function with 
// TagSize equal to sizeof(Tag), where Tag has been allocated as
// char Tag[something].  Note that this code won't work if the 
// quoted strings in the tag contain '>' or '<'.

static int
GetTag (FILE * fp, char *Tag, int TagSize)
{
  int ch;
  for (;;)
    {
      ch = getc (fp);
      if (ch == EOF)
	return (1);
      if (ch == '<')		// Found start of tag.
	{
	  for (;;)
	    {
	      ch = getc (fp);
	      if (ch == EOF)
		return (1);
	      if (ch == '>')	// Found end of tag.
		{
		  if (TagSize <= 0)
		    return (1);
		  *Tag = '\0';
		  return (0);
		}
	      else
		{
		  if (TagSize <= 0)
		    return (1);
		  *Tag++ = ch;
		  TagSize--;
		}
	    }
	}
    }
}

//------------------------------------------------------------------
// Read textual data between tags.

static int
GetBetweenTag (FILE * fp, char *Text, int TextSize)
{
  int ch;
  char c, *s;
  int i;
  s = Text;
  for (;;)
    {
      ch = getc (fp);
      if (ch == EOF)
	return (1);
      if (ch == '<')		// Found start of tag.
	{
	  ungetc (ch, fp);
	  if (TextSize <= 0)
	    return (1);
	  *Text = '\0';
	  break;
	}
      else
	{
	  if (TextSize <= 0)
	    return (1);
	  *Text++ = ch;
	  TextSize--;
	}
    }
  // Convert things like &#60; &#62; &#38; back to < > &.
  for (; *s; s++)
    if (2 == sscanf (s, "&#%d%c", &i, &c) && c == ';' &&
	((i >= ' ' && i < 127) || (i >= 160 && i <= 255)))
      {
	*s = i;
	strcpy (s + 1, s + 5 + (i > 99));
      }
  return (0);
}

//------------------------------------------------------------------
// Cleans up the text read from the input file in various ways, 
// such as trimming leading/trailing space and translating
// special characters like '<' and '>'.

static void
CleanupText (char *s)
{
  char *ss, *sss = NULL, c;
  for (ss = s; isspace (*ss); ss++);
  while (*ss)
    {
      if (*ss == '&')
	{
	  if (!strncasecmp (ss, "&amp;", 5) || !strncasecmp (ss, "&#38;", 5))
	    {
	      c = '&';
	      ss += 5;
	    }
	  else if (!strncasecmp (ss, "&gt;", 4))
	    {
	      c = '>';
	      ss += 4;
	    }
	  else if (!strncasecmp (ss, "&#62;", 5))
	    {
	      c = '>';
	      ss += 5;
	    }
	  else if (!strncasecmp (ss, "&lt;", 4))
	    {
	      c = '<';
	      ss += 4;
	    }
	  else if (!strncasecmp (ss, "&#60;", 5))
	    {
	      c = '<';
	      ss += 5;
	    }
	  else
	    c = *ss++;
	}
      else
	c = *ss++;
      *s++ = c;
      if (!isspace (c))
	sss = s;
    }
  if (sss == NULL)
    *s = '\0';
  else
    *sss = '\0';
}

//------------------------------------------------------------------
// Interprets the meanings of the applicability/cc fields in the tags.

static int
InterpretLevel (char *s)
{
  int RetVal;
  switch (*s)
    {
    case 'N':
      RetVal = SATISFIED_DISCRETION;
      break;
    case 'Y':
      RetVal = SATISFIED_YES;
      break;
    case 'I':
      RetVal = SATISFIED_INDEPENDENCE;
      break;
    default:
      return (0);
    }
  switch (s[1])
    {
    case 'N':
      RetVal |= CC_DISCRETION;
      break;
    case '1':
      RetVal |= CC_1;
      break;
    case '2':
      RetVal |= CC_2;
      break;
    default:
      return (0);
    }
  return (RetVal);
}

//---------------------------------------------------------------------
// Interprets the meanings of the document mappings.

static int
InterpretDoc (char *s)
{
  if (0 == strcasecmp (s, "YES"))
    return (DOCUMENT_YES);
  else
    return (DOCUMENT_DISCRETION);
}

//------------------------------------------------------------------
// Recursive function for adding trees gotten from the input XML
// file.  On the first entry, Level must be 0.  Returns 0 on success.
// Expects a tree to end with a tag like </something>, except if 
// Level == 0, in which case the end of file terminates.  Also, 
// Xml should already have been loaded with the root entry, and the
// opening tag for the root entry should have been removed from the
// input file. 

#define MAX_TAGSIZE 4096
#define MAX_TAGFIELDS 32
static char Tag[MAX_TAGSIZE];
static char TagFields[MAX_TAGFIELDS][MAX_TAGSIZE];
static char Text[32768];
static char ActualTag[128];

static int
LoadXmlLevel (int Level, FILE * fp, struct XmlStruct *Xml)
{
  struct XmlStruct *NewXml, Default = { NULL }, *Root;
  char *s, *ss;
  int ReturnValue = 1, i, Hinted;
  int Position, NumInLevel, TagFieldCount;
  Level++;
  Root = Xml;
  Default.Color = wxBLACK;	// Default text color black.
  for (NumInLevel = 0;; Xml = NewXml, NumInLevel++)
    {
      // Read the next tag.  If none found, it's an error unless
      // Level == 0, in which case it's a natural exit.
      if (GetTag (fp, Tag, sizeof (Tag)))
	{
	  // Maybe error exit, maybe not.
	  ReturnValue = 2 * (Level != 0);
	  goto Error;
	}
      // Okay, we've found a tag.  If it's an end-tag, return.
      if (Tag[0] == '/')
	goto Fine;
      // If this is the first tag we've found on this level, we need to
      // add a child to the root.  Otherwise, we need to add a sibling
      // to the last element added.
      NewXml = (struct XmlStruct *) malloc (sizeof (struct XmlStruct));
      if (NewXml == NULL)
	{
	  ReturnValue = 3;
	  goto Error;
	}
      *NewXml = Default;
      NewXml->Parent = Root;
      if (NumInLevel == 0)
	Xml->Children = NewXml;
      else
	Xml->Next = NewXml;
      // Parse the start-tag.  The format is this:
      //        <tag var1="data1" var2="data2" ... varN="dataN">
      // The "tag", "varN", and "dataN" strings do not contain spaces,
      // nor the characters '<', '>', or '"'.  In fact, all special 
      // characters other than '-' or alpha-numerics are represented by
      // numeric codes of the form "&#n;".  On input, '_' is also 
      // translated to ' '.  Here are the recognized variables:
      //        HINT, 
      //        TYPE,
      //        PSAC, SDP, ... (other docs) ..., SAS,
      //        A1, A2, A3, BA, BB, BC, BD, BE
      // Though technically any data strings meeting the conditions
      // described above are acceptable, the interpretations of the
      // variables actually limit what may be found:
      //        HINT    Arbitrary text.  Gives the user a hint as to
      //                what this particular data item is used for.
      //        TYPE    Limit to alphanumerics, '-', and '_'.
      //                This is by default the same as the tag, 
      //                but after user-editing the tag may change.
      //                The TYPE never changes.
      //        PSAC..SAS       These are used to determine whether 
      //                the particular type of data item appears
      //                in the indicated document type in general.
      //                In practice, whether it appears or not
      //                also depends on the criticality level.
      //                The allowed values are "NO" (the default)
      //                and "YES".
      //        A1..BE  These are used to the "applicability"
      //                and "control category" at the various
      //                criticality levels.  (Refer to DO-178B
      //                table A-1.)   The allowed values consist
      //                of two 1-character fields, the first 
      //                representing the "applicability":
      //                        N       "Applicant's discretion"
      //                        Y       Yes, applicable.
      //                        I       Yes, applicable with "independence".
      //                The second 1-character field determines
      //                the "control category":
      //                        N       "Applicant's discretion"
      //                        1       CC1
      //                        2       CC2
      TagFieldCount = sscanf (Tag, "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s"
			      "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
			      ActualTag,
			      TagFields[0], TagFields[1],
			      TagFields[2], TagFields[3],
			      TagFields[4], TagFields[5],
			      TagFields[6], TagFields[7],
			      TagFields[8], TagFields[9],
			      TagFields[10], TagFields[11],
			      TagFields[12], TagFields[13],
			      TagFields[14], TagFields[15],
			      TagFields[16], TagFields[17],
			      TagFields[18], TagFields[19],
			      TagFields[20], TagFields[21],
			      TagFields[22], TagFields[23],
			      TagFields[24], TagFields[25],
			      TagFields[26], TagFields[27],
			      TagFields[28], TagFields[29],
			      TagFields[30], TagFields[31]);
      NewXml->Type = (char *) malloc (strlen (ActualTag) + 1);
      if (NewXml->Type == NULL)
	{
	  ReturnValue = 4;
	  goto Error;
	}
      strcpy (NewXml->Type, ActualTag);
      TagToName (ActualTag);
      NewXml->Name = (char *) malloc (strlen (ActualTag) + 1);
      if (NewXml->Name == NULL)
	{
	  ReturnValue = 5;
	  goto Error;
	}
      strcpy (NewXml->Name, ActualTag);
      //  Analyze the tag's variables.
      TagFieldCount--;
      Hinted = 0;
      for (i = 0; i < TagFieldCount; i++)
	{
	  // Clean up the data portion.
	  for (s = TagFields[i]; *s && *s != '='; s++);
	  if (*s != '=' || s[1] != '\"')
	    continue;
	  s += 2;		// s points to start of the data.  
	  for (ss = s; *ss && *ss != '\"'; ss++);
	  if (*ss != '\"')
	    continue;
	  *ss = '\0';		// data area now nul-terminated.
	  // Now that the dat's swellerific, determine what 
	  // type of data it actually is!  In all cases except
	  // TYPE, the data is converted to the "non-tag" form.
	  if (0 == strncasecmp (TagFields[i], "TYPE=", 5))
	    {
	      free (NewXml->Type);	// Remove the default.
	      NewXml->Type = (char *) malloc (strlen (s) + 1);
	      if (NewXml->Type == NULL)
		{
		  ReturnValue = 7;
		  goto Error;
		}
	      strcpy (NewXml->Type, s);
	      continue;
	    }
	  TagToName (s);	// Clean up the data.  
	  if (0 == strncasecmp (TagFields[i], "HINT=", 5))
	    {
	      NewXml->Hint = (char *) malloc (strlen (s) + 1);
	      if (NewXml->Hint == NULL)
		{
		  ReturnValue = 6;
		  goto Error;
		}
	      strcpy (NewXml->Hint, s);
	      Hinted = 1;
	    }
	  else if (0 == strncasecmp (TagFields[i], "A1=", 3))
	    NewXml->Applicability[DO178A_LEVEL_1] = InterpretLevel (s);
	  else if (0 == strncasecmp (TagFields[i], "A2=", 3))
	    NewXml->Applicability[DO178A_LEVEL_2] = InterpretLevel (s);
	  else if (0 == strncasecmp (TagFields[i], "A3=", 3))
	    NewXml->Applicability[DO178A_LEVEL_3] = InterpretLevel (s);
	  else if (0 == strncasecmp (TagFields[i], "BA=", 3))
	    NewXml->Applicability[DO178B_LEVEL_A] = InterpretLevel (s);
	  else if (0 == strncasecmp (TagFields[i], "BB=", 3))
	    NewXml->Applicability[DO178B_LEVEL_B] = InterpretLevel (s);
	  else if (0 == strncasecmp (TagFields[i], "BC=", 3))
	    NewXml->Applicability[DO178B_LEVEL_C] = InterpretLevel (s);
	  else if (0 == strncasecmp (TagFields[i], "BD=", 3))
	    NewXml->Applicability[DO178B_LEVEL_D] = InterpretLevel (s);
	  else if (0 == strncasecmp (TagFields[i], "BE=", 3))
	    NewXml->Applicability[DO178B_LEVEL_E] = InterpretLevel (s);
	  else if (0 == strncasecmp (TagFields[i], "PSAC=", 5))
	    NewXml->Documents[PSAC] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "SDP=", 4))
	    NewXml->Documents[SDP] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "SVP=", 4))
	    NewXml->Documents[SVP] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "SCMP=", 5))
	    NewXml->Documents[SCMP] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "SQAP=", 5))
	    NewXml->Documents[SQAP] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "SRS=", 4))
	    NewXml->Documents[SRS] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "SDS=", 4))
	    NewXml->Documents[SDS] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "SCS=", 4))
	    NewXml->Documents[SCS] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "SRD=", 4))
	    NewXml->Documents[SRD] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "DD=", 3))
	    NewXml->Documents[DD] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "SVCP=", 5))
	    NewXml->Documents[SVCP] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "SECI=", 5))
	    NewXml->Documents[SECI] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "SCI=", 4))
	    NewXml->Documents[SCI] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "SAS=", 4))
	    NewXml->Documents[SAS] = InterpretDoc (s);
	}
      if (!Hinted)
	{
	  NewXml->Hint = (char *) malloc (strlen ("") + 1);
	  if (NewXml->Hint == NULL)
	    {
	      ReturnValue = 8;
	      goto Error;
	    }
	  strcpy (NewXml->Hint, "");
	}
      // Get the text, if any.
      if (GetBetweenTag (fp, Text, sizeof (Text)))
	{
	  ReturnValue = 9;
	  goto Error;
	}
      CleanupText (Text);
      NewXml->Text = (char *) malloc (strlen (Text) + 1);
      if (NewXml->Text == NULL)
	{
	  ReturnValue = 10;
	  goto Error;
	}
      strcpy (NewXml->Text, Text);
      // This must be followed either by the closing tag, or else by a
      // tag for a child-tree.  
      Position = ftell (fp);
      if (GetTag (fp, Tag, sizeof (Tag)))
	{
	  ReturnValue = 11;
	  goto Error;
	}
      // If it's a closing tag, and if it matches the opening tag,
      // then we're swell.
      if (Tag[0] == '/')
	{
	  TagToName (&Tag[1]);
	  if (0 == strcmp (&Tag[1], NewXml->Name))
	    continue;
	  else
	    {
	      ReturnValue = 12;
	      goto Error;
	    }
	}

      // If it's not a closing tag, then it must be an opening tag, and
      // we must do a child tree.
      fseek (fp, Position, SEEK_SET);
      if (LoadXmlLevel (Level, fp, NewXml))
	{
	  ReturnValue = 13;
	  goto Error;
	}
    }

  // All done!
Fine:
  ReturnValue = 0;
Error:
  Level--;
  return (ReturnValue);
}

//------------------------------------------------------------------
// Loads a Do178Builder XML file.
// Returns 0 on success.

int
LoadXmlFile (const char *Filename, struct XmlStruct *Xml)
{
  FILE *fp;
  int ReturnValue = 100;
  char *s;

  fp = fopen (Filename, "rb");
  if (fp == NULL)
    {
      ReturnValue = 101;
      goto Done;
    }
  ClearXml (Xml);

  // The recursive function that reads in the XML isn't going
  // to work unless the in-memory version already has a root
  // element, and the root has been removed from the file.
  // So let's do that.
  if (GetTag (fp, Tag, sizeof (Tag)))
    {
      ReturnValue = 102;
      goto Done;
    }
  for (s = Tag; *s && !isspace (*s); s++);
  *s = '\0';
  Xml->Type = (char *) malloc ((s - Tag) + 1);
  if (NULL == Xml->Type)
    {
      ReturnValue = 103;
      goto Done;
    }
  strcpy (Xml->Type, Tag);
  TagToName (Tag);
  Xml->Name = (char *) malloc (strlen (Tag) + 1);
  if (NULL == Xml->Name)
    {
      ReturnValue = 104;
      goto Done;
    }
  strcpy (Xml->Name, Tag);
  Xml->Text = (char *) malloc (1);
  if (NULL == Xml->Text)
    {
      ReturnValue = 105;
      goto Done;
    }
  strcpy (Xml->Text, "");
  Xml->Hint = (char *) malloc (1);
  if (NULL == Xml->Hint)
    {
      ReturnValue = 106;
      goto Done;
    }
  strcpy (Xml->Hint, "");

  // Now actually read the file.
  if (0 != (ReturnValue = LoadXmlLevel (0, fp, Xml)))
    goto Done;

  // This should have level the closing tag for the root level in the
  // buffer.  Let's check!
  if (Tag[0] != '/' || 0 != strcmp (&Tag[1], Xml->Type))
    {
      ReturnValue = 107;
      goto Done;
    }

  // All done!    
  ReturnValue = 0;
Done:
  if (fp != NULL)
    fclose (fp);
  if (ReturnValue != 0)
    ClearXml (Xml);
  return (ReturnValue);
}

//------------------------------------------------------------------
// Saves a single level of an XML structure to disk.  On entry,
// Xml points to the first element in the level.

// Names of the document types, and other variable names.
// These are just the various constants from Do178Builder.h,
// presented as string arrays.
static const char *DocVarNames[] = {
  "PSAC", "SDP", "SVP", "SCMP", "SQAP", "SRS", "SDS", "SCS",
  "SRD", "DD", "SVCP", "SECI", "SCI", "SAS"
};
static const char *LevelNames[] = {
  "A1", "A2", "A3", "BA", "BB", "BC", "BD", "BE"
};

static int
SaveXmlLevel (int Level, FILE * fp, struct XmlStruct *Xml)
{
  int RetVal = 1;
  int i;
  char *s;
  for (; Xml != NULL; Xml = Xml->Next)
    {
      // Indent.
      for (i = 0; i < Level; i++)
	putc (' ', fp);

      // Write the tag. 
      NameToTag (Xml->Name, Tag);
      fprintf (fp, "<%s", Tag);
      if (0 != strcmp (Xml->Type, Tag))
	{
	  fprintf (fp, " TYPE=\"%s\"", Xml->Type);
	}
      NameToTag (Xml->Hint, Tag);
      fprintf (fp, " HINT=\"%s\"", Tag);
      for (i = 0; i < NUM_DOCUMENTS; i++)
	if (Xml->Documents[i] != 0)
	  fprintf (fp, " %s=\"YES\"", DocVarNames[i]);
      for (i = 0; i < NUM_CRITICALITY_LEVELS; i++)
	if (Xml->Applicability[i] != 0)
	  {
	    fprintf (fp, " %s=\"", LevelNames[i]);
	    switch (Xml->Applicability[i] & SATISFIED_MASK)
	      {
	      case SATISFIED_DISCRETION:
		putc ('N', fp);
		break;
	      case SATISFIED_YES:
		putc ('Y', fp);
		break;
	      case SATISFIED_INDEPENDENCE:
		putc ('I', fp);
		break;
	      default:
		putc ('N', fp);
		break;
	      }
	    switch (Xml->Applicability[i] & CC_MASK)
	      {
	      case CC_DISCRETION:
		putc ('N', fp);
		break;
	      case CC_1:
		putc ('1', fp);
		break;
	      case CC_2:
		putc ('2', fp);
		break;
	      default:
		putc ('N', fp);
		break;
	      }
	    putc ('\"', fp);
	  }
      fprintf (fp, ">");

      // Write the text.
      for (s = Xml->Text; *s; s++)
	{
	  if (*s == '<')
	    fprintf (fp, "&lt;" /*"&#60;" */ );
	  else if (*s == '>')
	    fprintf (fp, "&gt;" /*"&#62;" */ );
	  else if (*s == '&')
	    fprintf (fp, "&amp;" /*"&#38;" */ );
	  else
	    putc (*s, fp);
	}

      // Write the child level, if any.
      if (Xml->Children != NULL)
	{
	  putc ('\n', fp);
	  if (SaveXmlLevel (Level + 1, fp, Xml->Children))
	    goto Done;
	}

      // Write the closing tag.
      NameToTag (Xml->Name, Tag);
      if (Xml->Children != NULL)
	for (i = 0; i < Level; i++)
	  putc (' ', fp);
      fprintf (fp, "</%s>\n", Tag);
    }

  RetVal = 0;
Done:
  return (RetVal);
}

//-----------------------------------------------------------------
// Check to see if a file exists.

int
FileExists (const char *Filename)
{
  FILE *fp;
  fp = fopen (Filename, "r");
  if (fp == NULL)
    return (0);
  fclose (fp);
  return (1);
}

//------------------------------------------------------------------
// Saves a Do178Builder XML file.
// Returns 0 on success.

int
SaveXmlFile (const char *Filename, struct XmlStruct *Xml)
{
  FILE *fp;
  int ReturnValue = 1;
  wxString TempFile, BackFile, DoubleBackFile;

  //fp = fopen (Filename, "wb");
  TempFile = Filename;
  TempFile = TempFile + ".tmp";
  BackFile = Filename;
  BackFile = BackFile + ".bak";
  DoubleBackFile = Filename;
  DoubleBackFile = DoubleBackFile + ".bk2";
  fp = fopen (TempFile, "wb");
  if (fp == NULL)
    goto Done;

  // We have a recursive function to save the file, 
  // because that allows us to descend to lower levels easily.
  if (SaveXmlLevel (0, fp, Xml))
    goto Done;

  // The file has now been created, as a temporary file.
  // We need to move the original to the backup, and the
  // temporary to the original. If it doesn't work, we
  // do our darndest to leave the original file and the
  // primary backup unchanged.  The secondary backup will
  // usually have been deleted.
  if (fclose (fp))
    goto Done;
  fp = NULL;
  remove (DoubleBackFile);
  if (FileExists (BackFile))
    {
      if (rename (BackFile, DoubleBackFile))
	{
	  goto Done;
	}
    }
  if (FileExists (Filename))
    {
      if (rename (Filename, BackFile))
	{
	  rename (DoubleBackFile, BackFile);
	  goto Done;
	}
    }
  if (rename (TempFile, Filename))
    {
      rename (BackFile, Filename);
      rename (DoubleBackFile, BackFile);
      goto Done;
    }

  // All done!    
  ReturnValue = 0;
Done:
  if (fp != NULL)
    fclose (fp);
  return (ReturnValue);
}

//------------------------------------------------------------------------
// Navigates through the XML linked list, finding items by numerical index
// rather than hierarchy.   

struct XmlStruct *
FindXml (struct XmlStruct *Root, wxTreeItemId Index)
{
  XmlStruct *Next, *RetVal;
  for (Next = Root; Next != NULL; Next = Next->Next)
    {
      if (Index == Next->Id)
	return (Next);
      if (Next->Children != NULL)
	{
	  RetVal = FindXml (Next->Children, Index);
	  if (RetVal != NULL)
	    return (RetVal);
	}
    }

  return (NULL);
}

//-------------------------------------------------------------------------
// Determine all of the section numbers.  Can operate recursively.  
// Basically, finds the section number of the closest ancestor relevant to
// the same document, and adds stuff onto that section number.
// On entry, Root is the root of a sub-tree, and Root itself has already 
// been processed, except when Root is itself the root of the entire tree.
//
// In assigning new numbers at a given level, we need to know what numbers
// have already been used at that level.  That's harder to do than it 
// might seem, because the outline structure of the printed documents can
// differ from that of the online data.  The quick&dirty solution is to 
// just keep an array of the used section numbers, and if we find that a 
// number we want to use is taken, we just bump it upward.

static wxStringList *UsedSections[NUM_DOCUMENTS];

void
HtmlSections (struct XmlStruct *Root)
{
  int i, SubSection[NUM_DOCUMENTS];
  struct XmlStruct *Parent, *Record;
  for (i = 0; i < NUM_DOCUMENTS; SubSection[i++] = 1);
  // Loop on the sub-items.
  for (Record = Root->Children; Record != NULL; Record = Record->Next)
    {
      // By default, clear all of the section numbers for this record.
      for (i = 0; i < NUM_DOCUMENTS; i++)
	{
	  *Record->Sections[i] = '\0';
	  Record->SectionLevel[i] = 0;
	}
      // If this record is still named TBDsomething, ignore it!
#ifdef AUTO_ELIMINATE_TBD
      if (0 == strncmp (Record->Name, "TBD", 3))
	continue;
#endif
      // For each document type, search the ancestors to find those
      // relevant to that document type.  The root entry and the 
      // Project_Overall_Information sub-item of the root don't count.
      // Also, any section title "Figure" is really a figure.
      if (0 == strcmp (Record->Type, "Project_Overall_Information"))
	continue;
      if (0 == strcmp (Record->Type, "Figure"))
	continue;
      for (i = 0; i < NUM_DOCUMENTS; i++)
	if (Record->Documents[i] != 0)
	  for (Parent = Record->Parent; Parent != NULL;
	       Parent = Parent->Parent)
	    if (Parent->Parent == NULL)
	      {
		Record->SectionLevel[i] = 0;
		do
		  sprintf (Record->Sections[i], "%d", SubSection[i]++);
		while (UsedSections[i]->Member (Record->Sections[i]));
		UsedSections[i]->Add (Record->Sections[i]);
		//Record->Sections[i] << SubSection[i]++;
	      }
	    else if (Parent->Documents[i] != 0)
	      {
		Record->SectionLevel[i] = Parent->SectionLevel[i] + 1;
		do
		  sprintf (Record->Sections[i], "%s.%d", Parent->Sections[i],
			   SubSection[i]++);
		while (UsedSections[i]->Member (Record->Sections[i]));
		UsedSections[i]->Add (Record->Sections[i]);
		//Record->Sections[i] << Parent->Sections[i] << "." << SubSection[i]++;
		break;
	      }
      // Take care of the subtree.
      HtmlSections (Record);
    }
}

//---------------------------------------------------------------------
// Writes an arbitrary text string as paragraph data.  In other words,
// as <para>%s</para><para>%s</para>... where each paragraph is initially
// delimited by newlines.  Returns the number of paragraphs written, 
// or zero on error.

// This array holds all of the inline DocBook tags we're willing to
// recognize.  (Actually, this list contains all of the inline elements
// that I know about from DocBook 3.1.)
const char *InlineTags[] = {
  //"Abbrev", 
  //"Accel",
  //"Acronym", 
  //"Action",
  //"Anchor",   // Requires an attribute.
  //"Application",
  "Citation",
  //"CiteRefEntry",
  "CiteTitle",
  "ClassName",
  "Command",
  "ComputerOutput",
  "Constant",
  "Database",
  //"Email",    // Seems not to be includable in <para>
  "Emphasis",
  "EnVar",
  "ErrorCode",
  "ErrorName",
  "ErrorType",
  "Filename",
  "FirstTerm",
  //"Footnote", // Must itself contain <para>
  "ForeignPhrase",
  "Function",
  //"GlossTerm",// "There is no glossary"
  //"GuiButton",
  //"GuiIcon",
  //"GuiLabel",
  //"GuiMenu",
  //"GuiMenuItem",
  //"GuiSubMenu",
  "Hardware",
  //"InlineEquation",
  //"InlineGraphic",
  //"Interface",
  //"InterfaceDefinition",
  "KeyCap",
  "KeyCode",
  //"KeyCombo",
  //"KeySym",
  //"Link",
  //"Literal",
  //"Markup",
  "MediaLabel",
  //"MenuChoice",
  "MouseButton",
  //"MsgText",
  //"OLink",
  "Option",
  "Optional",
  "Parameter",
  //"Phrase",
  "ProgramListing",
  "Prompt",
  "Property",
  //"Quote",
  "Replaceable",
  "ReturnValue",
  //"SgmlTag",
  //"Shortcut",
  "StructField",
  "StructName",
  "Subscript",
  "Superscript",
  "Symbol",
  //"SystemItem",
  "Token",
  "Trademark",
  "Type",
  //"ULink",
  "UserInput",
  "VarName"
    //"WordAsWord",
    //"XRef"
};
#define NUM_INLINETAGS (sizeof(InlineTags)/sizeof(const char *))
int NumAllInlines = NUM_INLINETAGS;
static int InlineLengths[NUM_INLINETAGS];
static int InlineLengthsPrepared = 0;
// This array holds just the "common" inlines suitable for novices.
const char *NoviceTags[] = {
  "CiteTitle",
  "Command",
  "Emphasis",
  "Function",
  "KeyCap",
  "Option",
  "Optional",
  "ProgramListing",
  "Subscript",
  "Superscript",
  "Trademark"
};
#define NUM_NOVICETAGS (sizeof(NoviceTags)/sizeof(const char *))
int NumNovices = NUM_NOVICETAGS;

static int
WriteParagraphs (FILE * fp, int IndentLevel, char *Text)
{
  int Paragraphs = 0;
  int ParagraphOpen = 0;
  int InProgramListing = 0;
  char *s, *ss;
  unsigned i;
  // Make sure that all of the InlineLengths are known.
  if (!InlineLengthsPrepared)
    {
      InlineLengthsPrepared = 1;
      for (i = 0; i < NUM_INLINETAGS; i++)
	InlineLengths[i] = strlen (InlineTags[i]);
    }
  // Start writing stuff!  
  if (*Text != '\0')
    {
      for (s = Text; *s;)
	{
	  if (*s == '\r')
	    {
	      s++;
	      continue;
	    }
	  // Allow <programlisting> markup to pass through.  
	  // This is treated separately from InlineTags, 
	  // because it can contain newlines.  
	  if (InProgramListing)
	    {
	      if (!strncasecmp (s, "</programlisting>", 17))
		{
		  InProgramListing = 0;
		  fprintf (fp, "</programlisting>");
		  s += 17;
		}
	      else
		{
		  putc (*s, fp);
		  s++;
		}
	      continue;
	    }
	  else if (!strncasecmp (s, "<programlisting>", 16))
	    {
	      InProgramListing = 1;
	      fprintf (fp, "<programlisting>");
	      s += 16;
	      continue;
	    }
	  if (*s == '\n')
	    {
	      if (ParagraphOpen)
		{
		  fprintf (fp, "</para>\n");
		  //if (EOF == fputs ("</para>\n", fp))
		  //  return (3);
		  //s++;
		  ParagraphOpen = 0;
		}
	      s++;
	      continue;
	    }
	  else if (!ParagraphOpen)
	    {
	      //for (i = 0; i < IndentLevel; i++)
	      //  if (EOF == putc (' ', fp))
	      //    return (3);
	      //if (EOF == fputs ("<P>", fp))
	      //  return (3);
	      fprintf (fp, "%*s<para>", IndentLevel + 1, "");
	      ParagraphOpen = 1;
	      Paragraphs++;
	    }
	  else if (*s == '<')
	    {
	      // We convert '<' to "&#60;" unless we can recognize
	      // it as inline markup. 
	      ss = s + 1 + (s[1] == '/');
	      for (i = 0; i < NUM_INLINETAGS; i++)
		if (!strncasecmp (ss, InlineTags[i], InlineLengths[i])
		    && ss[InlineLengths[i]] == '>')
		  {
		    ss += InlineLengths[i] + 1;
		    fwrite (s, ss - s, 1, fp);
		    s = ss;
		    break;
		  }
	      if (i < NUM_INLINETAGS)
		continue;
	      fprintf (fp, "&#60;");
	      s++;
	    }
	  else if (*s == '>')
	    {
	      fprintf (fp, "&#62;");
	      s++;
	    }
	  else if (*s == '&')
	    {
	      fprintf (fp, "&#38;");
	      s++;
	    }
	  else
	    putc (*s++, fp);
	}
      if (ParagraphOpen)
	{
	  if (EOF == fputs ("</para>\n", fp))
	    return (0);
	}
    }
  return (Paragraphs);
}

//-------------------------------------------------------------------------
// Output the section data for a specific SGML document.  Works recursively.
// On entry, Root should already have been processed.  We merely process
// all of the descendents.  Returns:
//      0               success
//      3               disk error
//      0x10            could not find source graphic
//      0x11            could not create destination graphic
//      0x12-0xff       other error

/*
This stuff was here from the olden days, when HTML was output 
rather than SGML.
#define MAX_HEADER_LEVELS 16
static const char *HeaderIn[MAX_HEADER_LEVELS] = {
  "<H1>", "<H2>", "<H3>", "<H4>",
  "<P><B>", "<P><B>", "<P><B>", "<P><B>",
  "<P><B>", "<P><B>", "<P><B>", "<P><B>",
  "<P><B>", "<P><B>", "<P><B>", "<P><B>"
};
static const char *HeaderOut[MAX_HEADER_LEVELS] = {
  "</H1>\n", "</H2>\n", "</H3>\n", "</H4>\n",
  "</B></P>\n", "</B></P>\n", "</B></P>\n", "</B></P>\n",
  "</B></P>\n", "</B></P>\n", "</B></P>\n", "</B></P>\n",
  "</B></P>\n", "</B></P>\n", "</B></P>\n", "</B></P>\n"
};
*/

static int
OutputSections (FILE * fp, int IndentLevel, int DocNum,
		struct XmlStruct *Root)
{
  struct XmlStruct *Next;
  int RetVal, ParagraphOpen, AnythingWritten, OpenSection;
  //char *s;
  //IndentLevel++;
  for (Next = Root->Children; Next != NULL; Next = Next->Next)
    {
#ifdef AUTO_ELIMINATE_TBD
      if (0 == strncmp (Next->Name, "TBD", 3))
	continue;
#endif
      OpenSection = 0;
      if (Next->Documents[DocNum] != 0)
	{
	  // If it's a figure, take care of it and continue.
	  if (0 == strcmp (Next->Type, "Figure"))
	    {
	      struct XmlStruct *Field, *FigField, *CapField;
	      // Find the filename:
	      FigField = NULL;
	      CapField = NULL;
	      for (Field = Next->Children; Field != NULL; Field = Field->Next)
		if (0 == strcmp (Field->Type, "Filename"))
		  FigField = Field;
		else if (0 == strcmp (Field->Type, "Caption"))
		  CapField = Field;
	      if (FigField == NULL)	// Malformed -- no Filename field.
		continue;
	      if (CapField != NULL)
		fprintf (fp, "%*s<figure><title>%s</title>", IndentLevel, "",
			 CapField->Text);
	      fprintf (fp,
		       "<graphic fileref=\"../%s\" align=\"center\"></graphic>",
		       FigField->Text);
	      if (CapField != NULL)
		fprintf (fp, "</figure>");
	      if (EOF == putc ('\n', fp))
		return (3);
	      continue;
	    }
	  // Write the section header.
	  OpenSection = 1;
	  fprintf (fp, "%*s<section>\n", IndentLevel, "");
	  fprintf (fp, "%*s<title>%s</title>\n", IndentLevel + 1, "",
		   Next->Name);
	  AnythingWritten = 0;
	  ParagraphOpen = 0;
	  // Write traceability data.  Traceability data is determined by the following
	  // method:  Firstly, the Type of the item must begin with TBD.  If the item
	  // is relevant to both SRD & SDD, or SAS & SVCP, or SDD & SVCP, then there is
	  // traceability between them.
	  if (strncmp (Next->Type, "TBD", 3) == 0)
	    {
	      if (Next->Documents[SVCP])
		{
		  if (Next->Documents[SDD]
		      && (DocNum == SVCP || DocNum == SDD))
		    {
		      // This is for test cases from detailed design.
		      fprintf (fp, "%*s<note><title>Traceability</title>",
			       IndentLevel + 1, "");
		      fprintf (fp,
			       "<para>SDD section %s &rarr; SVCP section %s.</para>",
			       Next->Sections[SDD], Next->Sections[SVCP]);
		      fprintf (fp, "</note>\n");
		    }
		  else if (Next->Documents[SAS]
			   && (DocNum == SVCP || DocNum == SAS))
		    {
		      // This is for test cases arising from changes since prior release.
		      fprintf (fp, "%*s<note><title>Traceability</title>",
			       IndentLevel + 1, "");
		      fprintf (fp,
			       "<para>SAS section %s &rarr; SVCP section %s.</para>",
			       Next->Sections[SAS], Next->Sections[SVCP]);
		      fprintf (fp, "</note>\n");
		    }
		}
	      if (Next->Documents[SRD] && Next->Documents[SDD]
		  && (DocNum == SRD || DocNum == SDD))
		{
		  // This is for design details.
		  fprintf (fp, "%*s<note><title>Traceability</title>",
			   IndentLevel + 1, "");
		  fprintf (fp,
			   "<para>SRD section %s &rarr; SDD section %s.</para>",
			   Next->Sections[SRD], Next->Sections[SDD]);
		  fprintf (fp, "</note>\n");
		}
	    }
	  // Write the section data.
	  //if (0 == WriteParagraphs (fp, IndentLevel, Next->Text))
	  //  if (EOF == fprintf (fp, "%*s<para>(TBD)</para>", IndentLevel, ""))
	  //    return (3);
	  WriteParagraphs (fp, IndentLevel, Next->Text);
	}
      RetVal = OutputSections (fp, IndentLevel + OpenSection, DocNum, Next);
      if (RetVal != 0)
	return (RetVal);
      if (OpenSection)
	fprintf (fp, "%*s</section>\n", IndentLevel, "");
    }
  return (0);
}

//-------------------------------------------------------------------------
// Generate DocBook SGML for a specific DO-178B document.  The numbering 
// of the docs is the same as the constants PSAC, SDP, SVP, etc.  Returns:
//      0       Success
//      1       Unknown doc type.
//      2       Cannot create file.
//      3       Disk-write error.

static int
MakeSgmlDocument (int DocNum, struct XmlStruct *Root,
		  const char *OutputDirectory)
{
  int ReturnValue;
  FILE *fp = NULL;
  char DocAcronym[129], DocName[64], Filename[1024];
  struct XmlStruct *Record, *Parent;
  char *Company_Name = NULL, *Project_Name = NULL, *Author = NULL,
    *Document_Version = NULL, *Publication_Date = NULL,
    *CopyrightYears = NULL, *CopyrightHolder = NULL,
    *Legal_Notice = NULL, *FigureFilename = NULL;
  // Determine the document name.
  switch (DocNum)
    {
    case PSAC:
      strcpy (DocName, "Plan for Software Aspects of Certification");
      strcpy (DocAcronym, "PSAC");
      break;
    case SDP:
      strcpy (DocName, "Software Development Plan");
      strcpy (DocAcronym, "SDP");
      break;
    case SVP:
      strcpy (DocName, "Software Verification Plan");
      strcpy (DocAcronym, "SVP");
      break;
    case SCMP:
      strcpy (DocName, "Software Configuration Management Plan");
      strcpy (DocAcronym, "SCMP");
      break;
    case SQAP:
      strcpy (DocName, "Software Quality Assurance Plan");
      strcpy (DocAcronym, "SQAP");
      break;
    case SRS:
      strcpy (DocName, "Software Requirements Standards");
      strcpy (DocAcronym, "SRS");
      break;
    case SDS:
      strcpy (DocName, "Software Design Standards");
      strcpy (DocAcronym, "SDS");
      break;
    case SCS:
      strcpy (DocName, "Software Code Standards");
      strcpy (DocAcronym, "SCS");
      break;
    case SRD:
      strcpy (DocName, "Software Requirements Data");
      strcpy (DocAcronym, "SRD");
      break;
    case DD:
      strcpy (DocName, "Software Design Description");
      strcpy (DocAcronym, "SDD");
      break;
    case SVCP:
      strcpy (DocName, "Software Verification Cases and Procedures");
      strcpy (DocAcronym, "SVCP");
      break;
    case SECI:
      strcpy (DocName, "Software Life Cycle Environment Configuration Index");
      strcpy (DocAcronym, "SECI");
      break;
    case SCI:
      strcpy (DocName, "Software Configuration Index");
      strcpy (DocAcronym, "SCI");
      break;
    case SAS:
      strcpy (DocName, "Software Accomplishment Summary");
      strcpy (DocAcronym, "SAS");
      break;
    default:
      ReturnValue = 1;
      goto Done;
    }

  // Create the SGML file.
  sprintf (Filename, "%s%s.sgml", OutputDirectory, DocAcronym);
  fp = fopen (Filename, "w");
  if (fp == NULL)
    {
      ReturnValue = 2;
      goto Done;
    }

  // Generate the article header.  This would include the title page.
  // [We use DocBook 3.1 (4.1 is actually the current version when this was
  // started) because 3.1 is the one covered in the edition of 
  // "DocBook: The Definitive Guide" that I have.]  
  // The first step is to parse the "Project Overall Information" section to
  // determine what actually goes into the header.
  printf ("Here %d!\n", DocNum);
  for (Record = Root->Children; Record != NULL; Record = Record->Next)
    if (0 == strcmp (Record->Type, "Project_Overall_Information")
	&& Record->Children != NULL)
      {
	for (Record = Record->Children; Record != NULL; Record = Record->Next)
	  {
	    if (0 == strcmp (Record->Type, "Company_Name") &&
		NULL != Record->Text && 0 != strncmp (Record->Text, "TBD", 3))
	      Company_Name = Record->Text;
	    else if (0 == strcmp (Record->Type, "Project_Name") &&
		     NULL != Record->Text &&
		     0 != strncmp (Record->Text, "TBD", 3))
	      Project_Name = Record->Text;
	    else if (0 == strcmp (Record->Type, "Author") &&
		     NULL != Record->Text &&
		     0 != strncmp (Record->Text, "TBD", 3))
	      Author = Record->Text;
	    else if (0 == strcmp (Record->Type, "Document_Version") &&
		     NULL != Record->Text &&
		     0 != strncmp (Record->Text, "TBD", 3))
	      Document_Version = Record->Text;
	    else if (0 == strcmp (Record->Type, "Publication_Date") &&
		     NULL != Record->Text &&
		     0 != strncmp (Record->Text, "TBD", 3))
	      Publication_Date = Record->Text;
	    else if (0 == strcmp (Record->Type, "Copyright")
		     && Record->Children != NULL)
	      {
		Parent = Record;
		for (Record = Record->Children; Record != NULL;
		     Record = Record->Next)
		  {
		    if (0 == strcmp (Record->Type, "Years") &&
			NULL != Record->Text &&
			0 != strncmp (Record->Text, "TBD", 3))
		      CopyrightYears = Record->Text;
		    else if (0 == strcmp (Record->Type, "Holder") &&
			     NULL != Record->Text &&
			     0 != strncmp (Record->Text, "TBD", 3))
		      CopyrightHolder = Record->Text;
		  }
		Record = Parent;
	      }
	    else if (0 == strcmp (Record->Type, "Legal_Notice") &&
		     NULL != Record->Text &&
		     0 != strncmp (Record->Text, "TBD", 3))
	      Legal_Notice = Record->Text;
	    else if (0 == strcmp (Record->Type, "Figure")
		     && Record->Children != NULL)
	      {
		Parent = Record;
		for (Record = Record->Children; Record != NULL;
		     Record = Record->Next)
		  {
		    if (0 == strcmp (Record->Type, "Filename") &&
			NULL != Record->Text &&
			0 != strncmp (Record->Text, "TBD", 3))
		      FigureFilename = Record->Text;
		  }
		Record = Parent;
	      }
	  }
	break;
      }
  printf ("Now!\n");
  fprintf (fp,
	   "<!DOCTYPE article PUBLIC \"-//OASIS//DTD DocBook V3.1//EN\">\n");
  fprintf (fp, "<article>\n");
  fprintf (fp, " <artheader>\n");
  if (Company_Name != NULL)
    fprintf (fp, "  <corpname>%s</corpname>\n", Company_Name);
  if (Project_Name != NULL)
    fprintf (fp, "  <productname>%s</productname>\n", Project_Name);
  fprintf (fp, "  <title>");
  if (Project_Name != NULL)
    fprintf (fp, "<productname>%s</productname>, ", Project_Name);
  fprintf (fp, "%s</title>\n", DocName);
  fprintf (fp, "  <titleabbrev>");
  if (Project_Name != NULL)
    fprintf (fp, "<productname>%s</productname>, ", Project_Name);
  fprintf (fp, "%s</titleabbrev>\n", DocAcronym);
  if (Author != NULL)
    fprintf (fp, "  <corpauthor>%s</corpauthor>\n", Author);
  if (Document_Version != NULL)
    fprintf (fp, "  <releaseinfo>%s</releaseinfo>\n", Document_Version);
  if (Publication_Date != NULL)
    fprintf (fp, "  <PubDate>%s</PubDate>\n", Publication_Date);
  if (CopyrightYears != NULL && CopyrightHolder != NULL)
    fprintf (fp,
	     "  <copyright><year>%s</year><holder>%s</holder></copyright>\n",
	     CopyrightYears, CopyrightHolder);
  if (Legal_Notice != NULL)
    {
      fprintf (fp, "  <legalnotice>\n");
      if (0 == WriteParagraphs (fp, 3, Legal_Notice))
	if (EOF == fprintf (fp, "   <para>(TBD)</para>"))
	  return (3);
      fprintf (fp, "  </legalnotice>\n");
    }
  if (FigureFilename != NULL)
    fprintf (fp, "  <graphic fileref=\"%s\"></graphic>\n", FigureFilename);
  fprintf (fp, " </artheader>\n");
  if (fputs ("\n", fp) == EOF)
    {
    DiskError:
      ReturnValue = 3;
      goto Done;
    }

  // Now output the sections.
  OutputSections (fp, 1, DocNum, Root);

  // Close the SGML file.
  if (fputs ("\n</article>\n", fp) == EOF)
    goto DiskError;

  ReturnValue = 0;
Done:
  if (fp != NULL)
    fclose (fp);
  return (ReturnValue);
}

//-------------------------------------------------------------------------
// Output SGML DO-178B documents.  Returns 0 on success.  Otherwise,
// returns the doc number in bits 8-15 and an error code in bits 0-7.

int
XmlToSgml (struct XmlStruct *Root, const char *OutputDirectory)
{
  int i, RetVal = 0;
  // First pass:  Determine all of the section numbers.
  for (i = 0; i < NUM_DOCUMENTS; i++)
    {
      UsedSections[i] = new wxStringList;
      UsedSections[i]->Clear ();
    }
  HtmlSections (Root);
  // Second pass:  Generate SGML for each of the documents.
  for (i = 0; i < NUM_DOCUMENTS; i++)
    {
      RetVal = MakeSgmlDocument (i, Root, OutputDirectory);
      if (RetVal != 0)
	{
	  RetVal |= (i << 8);
	  break;
	}
    }
  for (i = 0; i < NUM_DOCUMENTS; i++)
    delete UsedSections[i];
  return (RetVal);
}
